"""
aioevents module.
"""

from .decorator import event_listener
from .emitter import EventEmitter
from .event import BaseEvent

__all__ = [
    "EventEmitter",
    "BaseEvent",
    "event_listener",
]



import asyncio
import pytest

from aioevent import EventEmitter, BaseEvent


pytestmark = pytest.mark.asyncio


class ChildEvent(BaseEvent):
    pass


class ParentEvent(BaseEvent):
    pass


class Child(EventEmitter):
    def trigger_event(self):
        self.emit(ChildEvent())


class Parent(EventEmitter):
    def __init__(self, loop):
        super().__init__(loop=loop)
        self.child = Child(loop=loop)
        self.proxy(self.child)

    def trigger_event(self):
        self.emit(ParentEvent())


class GrandParent(EventEmitter):
    def __init__(self, loop):
        super().__init__(loop=loop)
        self.child = Parent(loop=loop)
        self.proxy(self.child)


async def test_child_proxy(event_loop: "asyncio.AbstractEventLoop"):
    child_event_future: "asyncio.Future" = event_loop.create_future()

    async def event_handler(event: BaseEvent):
        assert isinstance(event, ChildEvent)
        child_event_future.set_result(True)

    parent = Parent(event_loop)
    parent.listen(ChildEvent, event_handler)
    parent.child.trigger_event()

    assert await asyncio.wait_for(child_event_future, 1)


async def test_nested_listeners(event_loop: "asyncio.AbstractEventLoop"):
    parent_event_future: "asyncio.Future" = event_loop.create_future()
    child_event_future: "asyncio.Future" = event_loop.create_future()

    async def parent_event_handler(event: BaseEvent):
        assert isinstance(event, ChildEvent)
        parent_event_future.set_result(True)

    async def child_event_handler(event: BaseEvent):
        assert isinstance(event, ChildEvent)
        child_event_future.set_result(True)

    parent = Parent(event_loop)
    parent.child.listen(ChildEvent, child_event_handler)
    parent.listen(ChildEvent, parent_event_handler)
    parent.child.trigger_event()

    assert await asyncio.wait_for(child_event_future, 1)
    assert await asyncio.wait_for(parent_event_future, 1)



"""
Provides EventEmitter class.
"""

import asyncio
import functools
from typing import (
    Callable,
    Dict,
    List,
    Type,
    Union,
    Coroutine,
)
from .event import BaseEvent


# pylint: disable=invalid-name
EventCallbackType = Callable[[BaseEvent], Union[None, Coroutine[None, None, None]]]


class EventEmitter:
    """
    ABC for a class whose instances emit events.

    :ivar _event_listeners: A dictionary whose keys are subclasses of BaseEvent
            and whose values are lists of event handlers. Event handlers should
            be callables that accept a single argument: the event being emitted.
    """

    _event_listeners: Dict[Type[BaseEvent], List[EventCallbackType]]

    def __init__(
            self,
            *args,
            loop: "asyncio.AbstractEventLoop" = None,
            **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.event_loop = loop or asyncio.get_event_loop()
        self._event_listeners = {}

    def listen(self, event_type: Type[BaseEvent], callback: EventCallbackType):
        """
        Register a callback to be fired when an event is emitted.
        :param event_type: The type of event (subclass of :py:class:`BaseEvent`)
                to listen for.
        :param callback: The callback to trigger when the event is emitted;
                should accept a single parameter which is the instance of
                `event_type` that was emitted.
        :return:
        """
        if event_type not in self._event_listeners:
            self._event_listeners[event_type] = []
        self._event_listeners[event_type].append(callback)

    def emit(self, event: BaseEvent):
        """
        Emit an event.
        :param event: The event to be omitted.
        :return:
        """
        handlers = self._get_handlers_for_type(type(event))
        for handler in handlers:
            self._call_handler(handler, event)

    def proxy(self, emitter):
        """
        Proxy events from another emitter.

        Useful in a grandparent-parent-child type pattern where the grandparent
        cares about events emitted in the child.
        :param emitter:
        :return:
        """
        emitter.listen(BaseEvent, self._proxy_event)

    def _proxy_event(self, event):
        """
        Emit an event via proxy.
        :param event: The event to proxy.
        :return:
        """
        self.emit(event)

    def _get_handlers_for_type(
            self, event_type: Type,
    ) -> List[EventCallbackType]:
        """
        Get all handlers for an event type.

        This method will walk up the subclass tree so that (e.g.) a handle
        for `SomeEventType` will also handle events of type
        `SubclassOfSomeEventType`.
        :param event_type: The event type to find handlers for.
        :return: A list of event handlers.
        """
        handlers = []
        if not issubclass(event_type, BaseEvent):
            return []
        if event_type in self._event_listeners:
            handlers.extend(self._event_listeners[event_type])
        for event_supertype in event_type.__bases__:
            handlers.extend(self._get_handlers_for_type(event_supertype))
        return handlers

    def _call_handler(self, handler: EventCallbackType, event: BaseEvent):
        if asyncio.iscoroutinefunction(handler):
            asyncio.ensure_future(handler(event), loop=self.event_loop)
        else:
            self.event_loop.call_soon(functools.partial(handler, event))



from typing import Type

from .emitter import EventEmitter
from .event import BaseEvent


def event_listener(emitter: EventEmitter, event_type: Type[BaseEvent]):
    """
    A decorator factory for creating easy inline event listeners.

    ::
        my_emitter = EventEmitter()
        MyEvent = type("MyEvent", (BaseEvent, ), {})

        @event_listener(my_emitter, MyEvent)
        def on_my_event(event):
            print("Got MyEvent!")

        # Prints "Got MyEvent!" on next iteration of event loop.
        my_emitter.emit(MyEvent())

    :param emitter:
    :param event_type:
    :return:
    """
    def decorator(method):
        print("Decorator!")
        emitter.listen(event_type, method)
        return method
    return decorator



from setuptools import setup, find_packages

setup(
    name="aioevent",
    author="Travis DePrato",
    author_email="travisdeprato@gmail.com",
    version="1.1.0",
    description="Events for asyncio.",
    packages=find_packages("."),
    install_requires=[],
    python_requires=">=3.6",
)



from typing import TYPE_CHECKING, NamedTuple
import pytest
from aioevent import EventEmitter, BaseEvent, event_listener

if TYPE_CHECKING:
    from asyncio.events import AbstractEventLoop


pytestmark = pytest.mark.asyncio


class PayloadEvent(BaseEvent):
    def __init__(self, payload):
        super().__init__()
        self.payload = payload
    payload: str


async def test_sync_decorator_listens(event_loop: "AbstractEventLoop"):
    emitter = EventEmitter()
    payload_future = event_loop.create_future()

    @event_listener(emitter, PayloadEvent)
    def on_payload_event(event: "PayloadEvent"):
        print("Got payload event:", event.payload)
        payload_future.set_result(event.payload)

    emitter.emit(PayloadEvent("foo!"))
    assert (await payload_future) == "foo!"


async def test_async_decorator_listens(event_loop: "AbstractEventLoop"):
    emitter = EventEmitter()
    payload_future = event_loop.create_future()

    @event_listener(emitter, PayloadEvent)
    async def on_payload_event(event: "PayloadEvent"):
        print("Got payload event:", event.payload)
        payload_future.set_result(event.payload)

    emitter.emit(PayloadEvent("foo!"))
    assert (await payload_future) == "foo!"



import asyncio
import pytest

from aioevent import EventEmitter, BaseEvent


pytestmark = pytest.mark.asyncio


class ConnectionOpenedEvent(BaseEvent):
    pass


class ConnectionClosedEvent(BaseEvent):
    pass


class ConnectionMock(EventEmitter):
    def __init__(self, loop):
        super().__init__(loop=loop)
        self.status = "uninitialized"

    def open(self):
        self.status = "open"
        self.emit(ConnectionOpenedEvent())

    def close(self):
        self.status = "closed"
        self.emit(ConnectionClosedEvent())


async def test_basic(event_loop: "asyncio.AbstractEventLoop"):

    seen_opened = event_loop.create_future()
    seen_closed: "asyncio.Future" = event_loop.create_future()

    def on_open(event):
        assert isinstance(event, ConnectionOpenedEvent)
        nonlocal seen_opened
        seen_opened.set_result(True)

    async def on_close(event):
        assert isinstance(event, ConnectionClosedEvent)
        await asyncio.sleep(0.1)
        nonlocal seen_closed
        seen_closed.set_result(True)

    connection = ConnectionMock(event_loop)
    connection.listen(ConnectionOpenedEvent, on_open)
    connection.listen(ConnectionClosedEvent, on_close)

    connection.open()
    assert await asyncio.wait_for(seen_opened, 1) and not seen_closed.done()
    connection.close()
    assert await asyncio.wait_for(seen_closed, 1) and seen_closed.result()


async def test_broad_listener(event_loop: "asyncio.AbstractEventLoop"):
    events_seen_future: "asyncio.Future" = event_loop.create_future()

    def on_event(event):
        nonlocal events_seen_future
        events_seen_future.set_result(True)

    connection = ConnectionMock(event_loop)
    connection.listen(BaseEvent, on_event)

    connection.open()
    assert await asyncio.wait_for(events_seen_future, 1)

    # reset future
    events_seen_future = event_loop.create_future()
    connection.close()
    assert await asyncio.wait_for(events_seen_future, 1)
